"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var TypeAttributes_1 = require("./TypeAttributes");
var MIN_LENGTH_FOR_ENUM = 10;
function shouldBeEnum(t) {
    var enumCases = t.enumCases;
    if (enumCases !== undefined) {
        Support_1.assert(enumCases.size > 0, "How did we end up with zero enum cases?");
        var numValues = enumCases.map(function (n) { return n; }).reduce(function (a, b) { return a + b; });
        if (numValues >= MIN_LENGTH_FOR_ENUM && enumCases.size < Math.sqrt(numValues)) {
            return t.enumCases;
        }
    }
    return undefined;
}
function replaceString(group, builder, forwardingRef) {
    Support_1.assert(group.size === 1);
    var t = Support_1.defined(group.first());
    var attributes = t.getAttributes();
    var maybeEnumCases = shouldBeEnum(t);
    if (maybeEnumCases !== undefined) {
        return builder.getEnumType(attributes, maybeEnumCases.keySeq().toOrderedSet(), forwardingRef);
    }
    return builder.getStringType(attributes, undefined, forwardingRef);
}
// A union needs replacing if it contains more than one string type, one of them being
// a basic string type.
function unionNeedsReplacing(u) {
    var stringMembers = u.stringTypeMembers;
    if (stringMembers.size <= 1)
        return undefined;
    if (u.findMember("string") === undefined)
        return undefined;
    return stringMembers;
}
// Replaces all string types in an enum with the basic string type.
function replaceUnion(group, builder, forwardingRef) {
    Support_1.assert(group.size === 1);
    var u = Support_1.defined(group.first());
    var stringMembers = Support_1.defined(unionNeedsReplacing(u));
    var stringAttributes = TypeAttributes_1.combineTypeAttributes(stringMembers.toArray().map(function (t) { return t.getAttributes(); }));
    var types = [];
    u.members.forEach(function (t) {
        if (stringMembers.has(t))
            return;
        types.push(builder.reconstituteType(t));
    });
    if (types.length === 0) {
        return builder.getStringType(TypeAttributes_1.combineTypeAttributes(stringAttributes, u.getAttributes()), undefined, forwardingRef);
    }
    types.push(builder.getStringType(stringAttributes, undefined));
    return builder.getUnionType(u.getAttributes(), immutable_1.OrderedSet(types), forwardingRef);
}
function inferEnums(graph, stringTypeMapping) {
    var allStrings = graph
        .allTypesUnordered()
        .filter(function (t) { return t instanceof Type_1.StringType; })
        .map(function (t) { return [t]; })
        .toArray();
    return graph.rewrite("infer enums", stringTypeMapping, false, allStrings, replaceString);
}
exports.inferEnums = inferEnums;
function flattenStrings(graph, stringTypeMapping) {
    var allUnions = graph.allNamedTypesSeparated().unions;
    var unionsToReplace = allUnions
        .filter(unionNeedsReplacing)
        .map(function (t) { return [t]; })
        .toArray();
    return graph.rewrite("flatten strings", stringTypeMapping, false, unionsToReplace, replaceUnion);
}
exports.flattenStrings = flattenStrings;
